"""Submodel for people.py"""
from typing import Optional

from pydantic import BaseModel


class Birthday(BaseModel):
    """Birthday model class.

    Used to represent birthday of person.
    """
    day: Optional[int]
    month: Optional[int]
    year: Optional[int]



"""Model for api/animes|mangas|ranobe/:id/roles"""
from typing import List, Optional

from pydantic import BaseModel

from .character import Character
from .person import Person


class Role(BaseModel):
    """Represents role info entity."""
    roles: List[str]
    roles_russian: List[str]
    character: Optional[Character]
    person: Optional[Person]



"""Model for /api/mangas"""
from typing import List, Optional

from pydantic import BaseModel, validator

from .genre import Genre
from .image import Image
from .publisher import Publisher
from .user_rate import UserRate
from .user_rate_score import UserRateScore
from .user_rate_status import UserRateStatus

MANGAS_KIND = ['manga', 'manhwa', 'manhua', 'one_shot', 'doujin']


class Manga(BaseModel):
    """Represents manga entity."""
    id: int
    name: str
    russian: str
    image: Image
    url: str
    kind: str
    score: float
    status: str
    volumes: int
    chapters: int
    aired_on: Optional[str]
    released_on: Optional[str]
    english: Optional[List[Optional[str]]]
    japanese: Optional[List[Optional[str]]]
    synonyms: Optional[List[Optional[str]]]
    license_name_ru: Optional[str]
    description: Optional[str]
    description_html: Optional[str]
    description_source: Optional[str]
    franchise: Optional[str]
    favoured: Optional[bool]
    anons: Optional[bool]
    ongoing: Optional[bool]
    thread_id: Optional[int]
    topic_id: Optional[int]
    myanimelist_id: Optional[int]
    rates_scores_stats: Optional[List[UserRateScore]]
    rates_statuses_stats: Optional[List[UserRateStatus]]
    licensors: Optional[List[str]]
    genres: Optional[List[Genre]]
    publishers: Optional[List[Publisher]]
    user_rate: Optional[UserRate]

    # pylint: disable=E0213
    @validator('kind')
    def kind_validator(cls, v):
        if v not in MANGAS_KIND:
            raise ValueError(f'Invalid manga kind. Got "{v}"'
                             f' but expected one of {MANGAS_KIND}')
        return v



"""Represents /api/ranobes resource."""
from typing import Any, Dict, List, Optional, Union, cast

from ..decorators import exceptions_handler, method_endpoint
from ..enums import RanobeCensorship, RanobeList, RanobeOrder, RanobeStatus
from ..exceptions import ShikimoriAPIResponseError
from ..models import FranchiseTree, Link, Ranobe, Relation, Role, Topic
from ..utils import Utils
from .base_resource import BaseResource


class Ranobes(BaseResource):
    """Ranobes resource class.

    Used to represent /api/ranobes resource.
    """

    @method_endpoint('/api/ranobe')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def get_all(self,
                      page: Optional[int] = None,
                      limit: Optional[int] = None,
                      order: Optional[RanobeOrder] = None,
                      status: Optional[Union[RanobeStatus,
                                             List[RanobeStatus]]] = None,
                      season: Optional[Union[str, List[str]]] = None,
                      score: Optional[int] = None,
                      genre: Optional[Union[int, List[int]]] = None,
                      publisher: Optional[Union[int, List[int]]] = None,
                      franchise: Optional[Union[int, List[int]]] = None,
                      censored: Optional[RanobeCensorship] = None,
                      my_list: Optional[Union[RanobeList,
                                              List[RanobeList]]] = None,
                      ids: Optional[Union[int, List[int]]] = None,
                      exclude_ids: Optional[Union[int, List[int]]] = None,
                      search: Optional[str] = None):
        """Returns ranobe list.

        :param page: Number of page
        :type page: Optional[int]

        :param limit: Number of results limit
        :type limit: Optional[int]

        :param order: Type of order in list
        :type order: Optional[RanobeOrder]

        :param status: Type(s) of ranobe status
        :type status: Optional[Union[RanobeStatus, List[RanobeStatus]]]

        :param season: Name(s) of ranobe seasons
        :type season: Optional[Union[str, List[str]]]

        :param score: Minimal ranobe score
        :type score: Optional[int]

        :param publisher: Publisher(s) ID
        :type publisher: Optional[Union[int, List[int]]

        :param genre: Genre(s) ID
        :type genre: Optional[Union[int, List[int]]

        :param franchise: Franchise(s) ID
        :type franchise: Optional[Union[int, List[int]]

        :param censored: Type of ranobe censorship
        :type censored: Optional[RanobeCensorship]

        :param my_list: Status(-es) of ranobe in current user list.
            If app is in restricted mode,
            this parameter won't affect on response.
        :type my_list: Optional[Union[RanobeList, List[RanobeList]]]

        :param ids: Ranobe(s) ID to include
        :type ids: Optional[Union[int, List[int]]

        :param exclude_ids: Ranobe(s) ID to exclude
        :type exclude_ids: Optional[Union[int, List[int]]

        :param search: Search phrase to filter ranobe by name
        :type search: Optional[str]

        :return: List of Ranobe
        :rtype: List[Ranobe]
        """
        query_dict = Utils.create_query_dict(page=page,
                                             limit=limit,
                                             order=order,
                                             status=status,
                                             season=season,
                                             score=score,
                                             genre=genre,
                                             publisher=publisher,
                                             franchise=franchise,
                                             censored=censored,
                                             mylist=my_list,
                                             ids=ids,
                                             exclude_ids=exclude_ids,
                                             search=search)

        response = await self._client.request(self._client.endpoints.ranobes,
                                              query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Ranobe)

    @method_endpoint('/api/ranobe/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def get(self, ranobe_id: int):
        """Returns info about certain ranobe.

        :param ranobe_id: Ranobe ID to get info
        :type ranobe_id: int

        :return: Ranobe info
        :rtype: Optional[Ranobe]
        """
        response = await self._client.request(
            self._client.endpoints.ranobe(ranobe_id))

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Ranobe)

    @method_endpoint('/api/ranobe/:id/roles')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def roles(self, ranobe_id: int):
        """Returns roles info of certain ranobe.

        :param ranobe_id: Ranobe ID to get creators
        :type ranobe_id: int

        :return: List of ranobe roles
        :rtype: List[Role]
        """
        response = await self._client.request(
            self._client.endpoints.ranobe_roles(ranobe_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Role)

    @method_endpoint('/api/ranobe/:id/similar')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def similar(self, ranobe_id: int):
        """Returns list of similar ranobes for certain ranobe.

        :param ranobe_id: Ranobe ID to get similar ranobes
        :type ranobe_id: int

        :return: List of similar ranobes
        :rtype: List[Ranobe]
        """
        response = await self._client.request(
            self._client.endpoints.similar_ranobes(ranobe_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Ranobe)

    @method_endpoint('/api/ranobe/:id/related')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def related(self, ranobe_id: int):
        """Returns list of related content of certain ranobe.

        :param ranobe_id: Ranobe ID to get related content
        :type ranobe_id: int

        :return: List of relations
        :rtype: List[Relation]
        """
        response = await self._client.request(
            self._client.endpoints.ranobe_related_content(ranobe_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Relation)

    @method_endpoint('/api/ranobe/:id/franchise')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def franchise(self, ranobe_id: int):
        """Returns franchise tree of certain ranobe.

        :param ranobe_id: Ranobe ID to get franchise tree
        :type ranobe_id: int

        :return: Franchise tree of certain ranobe
        :rtype: Optional[FranchiseTree]
        """
        response = await self._client.request(
            self._client.endpoints.ranobe_franchise_tree(ranobe_id))

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=FranchiseTree)

    @method_endpoint('/api/ranobe/:id/external_links')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def external_links(self, ranobe_id: int):
        """Returns list of external links of certain ranobe.

        :param ranobe_id: Ranobe ID to get external links
        :type ranobe_id: int

        :return: List of external links
        :rtype: List[Link]
        """
        response = await self._client.request(
            self._client.endpoints.ranobe_external_links(ranobe_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Link)

    @method_endpoint('/api/ranobe/:id/topics')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def topics(self,
                     ranobe_id: int,
                     page: Optional[int] = None,
                     limit: Optional[int] = None):
        """Returns list of topics of certain ranobe.

        :param ranobe_id: Ranobe ID to get topics
        :type ranobe_id: int

        :param page: Number of page
        :type page: Optional[int]

        :param limit: Number of results limit
        :type limit: Optional[int]

        :return: List of topics
        :rtype: List[Topic]
        """
        query_dict = Utils.create_query_dict(page=page, limit=limit)

        response = await self._client.request(
            self._client.endpoints.ranobe_topics(ranobe_id), query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Topic)



"""Model for /api/animes|mangas|ranobe/:id/related"""
from typing import Optional, Union

from pydantic import BaseModel

from .anime import Anime
from .manga import Manga
from .ranobe import Ranobe


class Relation(BaseModel):
    """Represents relation entity."""
    relation: str
    relation_russian: str
    anime: Optional[Anime]
    manga: Optional[Union[Manga, Ranobe]]



"""Model for /api/users/:id/history"""
from datetime import datetime
from typing import Optional, Union

from pydantic import BaseModel

from .anime import Anime
from .manga import Manga
from .ranobe import Ranobe


class History(BaseModel):
    """Represents user history timeline entity."""
    id: int
    created_at: datetime
    description: str
    target: Optional[Union[Anime, Manga, Ranobe]]



"""Model for /api/animes"""
from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel

from .genre import Genre
from .image import Image
from .screenshot import Screenshot
from .studio import Studio
from .user_rate import UserRate
from .user_rate_score import UserRateScore
from .user_rate_status import UserRateStatus
from .video import Video


class Anime(BaseModel):
    """Represents an anime entity."""
    id: int
    name: str
    russian: str
    image: Image
    url: str
    kind: Optional[str]
    score: float
    status: str
    episodes: int
    episodes_aired: int
    aired_on: Optional[str]
    released_on: Optional[str]
    rating: Optional[str]
    english: Optional[List[Optional[str]]]
    japanese: Optional[List[Optional[str]]]
    synonyms: Optional[List[Optional[str]]]
    license_name_ru: Optional[str]
    duration: Optional[int]
    description: Optional[str]
    description_html: Optional[str]
    description_source: Optional[str]
    franchise: Optional[str]
    favoured: Optional[bool]
    anons: Optional[bool]
    ongoing: Optional[bool]
    thread_id: Optional[int]
    topic_id: Optional[int]
    myanimelist_id: Optional[int]
    rates_scores_stats: Optional[List[UserRateScore]]
    rates_statuses_stats: Optional[List[UserRateStatus]]
    updated_at: Optional[datetime]
    next_episode_at: Optional[datetime]
    fansubbers: Optional[List[str]]
    fandubbers: Optional[List[str]]
    licensors: Optional[List[str]]
    genres: Optional[List[Genre]]
    studios: Optional[List[Studio]]
    videos: Optional[List[Video]]
    screenshots: Optional[List[Screenshot]]
    user_rate: Optional[UserRate]
    roles: Optional[List[str]]
    role: Optional[str]



"""Model for /api/people and submodel for creator.py"""
from datetime import datetime
from typing import List, Optional, Tuple

from pydantic import BaseModel

from .birthday import Birthday
from .image import Image
from .roles import Roles
from .works import Works


class Person(BaseModel):
    """Represents person entity."""
    id: int
    name: str
    russian: str
    image: Image
    url: str
    japanese: Optional[str]
    job_title: Optional[str]
    birthday: Optional[Birthday]
    website: Optional[str]
    groupped_roles: Optional[List[Tuple[str, int]]]
    roles: Optional[List[Roles]]
    works: Optional[List[Works]]
    thread_id: Optional[int]
    topic_id: Optional[int]
    person_favoured: Optional[bool]
    producer: Optional[bool]
    producer_favoured: Optional[bool]
    mangaka: Optional[bool]
    mangaka_favoured: Optional[bool]
    seyu: Optional[bool]
    seyu_favoured: Optional[bool]
    updated_at: Optional[datetime]



"""Submodel for creator.py"""
from datetime import datetime
from typing import List, Optional, Union

from pydantic import BaseModel

from .anime import Anime
from .image import Image
from .manga import Manga
from .ranobe import Ranobe
from .seyu import Seyu


class Character(BaseModel):
    """Represents character entity."""
    id: int
    name: str
    russian: str
    image: Image
    url: str
    altname: Optional[str]
    japanese: Optional[str]
    description: Optional[str]
    description_html: Optional[str]
    description_source: Optional[str]
    favoured: Optional[bool]
    thread_id: Optional[int]
    topic_id: Optional[int]
    updated_at: Optional[datetime]
    seyu: Optional[List[Seyu]]
    animes: Optional[List[Anime]]
    mangas: Optional[List[Union[Manga, Ranobe]]]



"""Model for /api/users/:id/anime_rates | manga_rates"""
from datetime import datetime
from typing import Optional, Union

from pydantic import BaseModel

from .anime import Anime
from .manga import Manga
from .ranobe import Ranobe
from .user import User


class UserList(BaseModel):
    """Represents user list entity.

    Contains data of watched/read titles.
    """
    id: int
    score: int
    status: str
    text: Optional[str]
    episodes: Optional[int]
    chapters: Optional[int]
    volumes: Optional[int]
    text_html: str
    rewatches: int
    created_at: datetime
    updated_at: datetime
    user: Optional[User]
    anime: Optional[Anime]
    manga: Optional[Union[Manga, Ranobe]]
